"""Package for analysis of magnetic structures."""

from __future__ import annotations

from pymatgen.analysis.magnetism.analyzer import (
    CollinearMagneticStructureAnalyzer,
    MagneticStructureEnumerator,
    Ordering,
    OverwriteMagmomMode,
    magnetic_deformation,
)
